import java.lang.*;
import java.io.*;

/*
Si consideri la definizione di coppia in Pair.Java, per coppie modificabili.
Si chiede di:
 (a) Ridefinirla in un un ADT, ImPairADTX.java, per coppie non modificabili e
     equipaggiata degli additionals (equals,clone,toString,elements) necessari;
 (b) Fornire AF e I  di (della rappresentazione data in) PairADT.java;
 (c) Fornire un caso di uso di PairADT.java;
 (d) Provare la correttezza di I.

 public class Pair <A,B> {
   public A left;
   public B right;
   public Pair (A x, B y) {
      left = x;
      right = y;
   }
}
*/

/* (a) */
interface MuCloneable extends Cloneable{
    public Object clone()throws CloneNotSupportedException;
}

class invalidArgException extends Exception{
}

class MuPairADTX <A,B> implements MuCloneable{
    private A left;//da nascondere
    private B right;//da nascondere
    
    public MuPairADTX(A x, B y) throws invalidArgException{
        if(x == null || y==null) throw new invalidArgException();
        left = x;
        right = y;
    }
    public A getLeft(){//da aggiungere per uso effettivo
        return left;
    }
    public B getRight(){//da aggiungere per uso effettivo
        return right;
    }
    public void setLeft(A x)throws invalidArgException{
        //da aggiungere per uso effettivo
        if(x==null) throw new invalidArgException();
        left = x;
    }
    public void setRight(B y)throws invalidArgException{
        //da aggiungere per uso effettivo
        if(y==null) throw new invalidArgException();
        right = y;
    }
    //public boolean equals(Object o)//ereditata
    public MuPairADTX<A,B> clone() throws CloneNotSupportedException{
            //dobbiamo catturare l'eccezione sollevabile dal costruttore quando
            //è applicato ad argomenti null.
        MuPairADTX<A,B> res = null;
        try{
            A newLeft  =left;
            B newRight = right;
            if (left instanceof MuCloneable)newLeft = (A)((MuCloneable)left).clone();
            if (right instanceof MuCloneable)newRight = (B)((MuCloneable)right).clone();
            res = new MuPairADTX<A,B>(newLeft,newRight);
        }
        catch(invalidArgException e){return null;}
        return res;
    }
    public String toString(){
        return ("("+left.toString()+","+right.toString()+")");
    }
}

/* (b)
 Sia [ImPairADT<Ta,Tb>]={(x,y) | x\inTa, y\inTb, Ta\in[A], Tb\in[B]}, l'insieme dei valori
 astratti definiti da ImPairADT<Ta,Tb>. Allora, per ogni valore concreto c esprimibile con
 lo stato concreto definito dai fields della classe:
 AF(c) = (c.left,c.right)
 I(c) = c.left!=null & c.right!=null
 */

/* (c) */
class ref<T> implements MuCloneable{
    T val;
    public ref(T n){
        val = n;
    }
    public boolean equals(Object o){//override equals
        ref<T> ok;
        try{ok = (ref<T>)o;}
        catch(Exception e){return false;}
        return (val.equals(ok.val));
    }
    public ref<T> clone(){
        return new ref(val);
    }
    public String toString(){
        return (val.toString());
    }
}

class main{
    public static void main(String args[]) throws invalidArgException,
            CloneNotSupportedException{//carte "truccate" sul valore
            ref<Integer> due = new ref<Integer>(2);
            ref<Integer> ace = new ref<Integer>(1);
            /* usiamo le carte introdotte */
            MuPairADTX<ref<Integer>,String> twoofclubs = new MuPairADTX<ref<Integer>,String>(due,"fiori");
            System.out.println("il valore della carta è " + twoofclubs.toString());
            MuPairADTX<ref<Integer>,String> myClone = (MuPairADTX <ref<Integer>,String>)(twoofclubs.clone());
            System.out.println("la carta e il suo clone: carta: " + twoofclubs.toString() + " ; clone: " + myClone.toString());
            System.out.println("stessa carta? " + (twoofclubs.equals(myClone)?"SI":"NO"));
            myClone.setLeft(ace);
            System.out.println("la carta e il suo clone: carta: " + twoofclubs.toString() + " ; clone: " + myClone.toString());
    }
}

/* (d)
 Dobbiamo provare che I(c)= c.left!=null & c.right!=null è sempre verificato dallo stato
 di un valore astratto di tipo MuPairADT.
 Fatto1: Costruttori. Il primo stm controlla che il valore (x,y) è costruito solo se x!=null!=y.
 Fatto2: Modificatori. Dal primo stm di setLeft (risp. setRight) vediamo che lo stato è modificato solo se l'argomento è diverso da null. Allora assumendo come ipotesi induttiva l'invariante vero sullo stato prima dell'invocazione, concludiamo che l'invariante è vero dopo un qualunque numero di invocazioni dei modificatori.
 Conclusione. Da Fatto1 e Fatto2 segue che I(c) è sempre soddisfatto.
 */

